import logging
import pyaml
import typing as T

from bdb import BdbQuit
from chainer.serializers import save_npz
from chainer.training import extensions
from chainer.training import updaters
from cvdatasets.utils import pretty_print_dict
from pathlib import Path


from cvfinetune.finetuner.mixins.base import BaseMixin
from cvfinetune.training.extensions import SacredReport
from cvfinetune.training.extensions import ManualGCCollect
from cvfinetune.utils.sacred import Experiment


class _TrainerMixin(BaseMixin):
    """This mixin is responsible for updater, evaluator and trainer creation.
    Furthermore, it implements the run method
    """

    def __init__(self, *args,
                 updater_cls=updaters.StandardUpdater,
                 updater_kwargs: dict = {},

                 only_eval: bool = False,
                 init_eval: bool = False,

                 experiment_name: T.Optional[str] = None,
                 no_snapshot: bool = False,
                 no_sacred: bool = False,

                 manual_gc: bool = True,
                 **kwargs):
        super(_TrainerMixin, self).__init__(*args, **kwargs)
        self.updater_cls = updater_cls
        self.updater_kwargs = updater_kwargs

        self.only_eval = only_eval
        self.init_eval = init_eval
        self.no_snapshot = no_snapshot
        self.no_sacred = no_sacred
        self.experiment_name = experiment_name
        self.manual_gc = manual_gc

        self.ex = None

    @property
    def no_observe(self):
        return self.no_sacred

    def init_experiment(self, *, config: dict):
        """ creates a sacred experiment that is later used
            by the trainer's sacred extension
        """

        self.config = config

        if self.no_sacred:
            logging.warning("Default sacred workflow is disabled "\
                "by the --no_sacred option!")
            return

        self.ex = Experiment(
            name=self.experiment_name,
            config=self.config,
            no_observe=self.no_observe)

        # self.trainer will be initialized later
        def run(*args, **kwargs):
            self._check_attr("trainer")
            return self.trainer.run(*args, **kwargs)

        self.ex.main(run)

    def run_experiment(self, *args, **kwargs):

        if self.ex is None:
            return self.trainer.run(*args, **kwargs)

        sacred_reporter = SacredReport(ex=self.ex, trigger=(1, "epoch"))
        self.trainer.extend(sacred_reporter)
        return self.ex(*args, **kwargs)


    def init_updater(self):
        """Creates an updater from training iterator and the optimizer."""

        self._check_attr("opt")
        self._check_attr("device")
        self._check_attr("train_iter")

        if self.opt is None:
            self.updater = None
            return

        self.updater = self.updater_cls(
            iterator=self.train_iter,
            optimizer=self.opt,
            device=self.device,
            **self.updater_kwargs,
        )
        logging.info(" ".join([
            f"Using single GPU: {self.device}.",
            f"{self.updater_cls.__name__} is initialized",
            f"with following kwargs: {pretty_print_dict(self.updater_kwargs)}"
            ])
        )

    def init_evaluator(self, default_name="val"):
        """Creates evaluation extension from validation iterator and the classifier."""

        self._check_attr("device")
        self._check_attr("val_iter")

        self.evaluator = extensions.Evaluator(
            iterator=self.val_iter,
            target=self.clf,
            device=self.device,
            progress_bar=True
        )

        self.evaluator.default_name = default_name

    def _new_trainer(self, trainer_cls, opts, *args, **kwargs):
        return trainer_cls(
            opts=opts,
            updater=self.updater,
            evaluator=self.evaluator,
            *args, **kwargs
        )

    def run(self, trainer_cls, opts, *args, **kwargs):

        self.trainer = self._new_trainer(trainer_cls, opts, *args, **kwargs)

        if self.manual_gc:
            manual_gc = ManualGCCollect(trigger=(1, "iteration"))
            self.trainer.extend(manual_gc)

        self.save_meta_info()

        logging.info("Snapshotting is {}abled".format(
            "dis" if self.no_snapshot else "en"))

        try:
            self.run_experiment(self.init_eval or self.only_eval)
        except (KeyboardInterrupt, BdbQuit) as e:
            raise e
        except Exception as e:
            self.dump("exception")
            raise e
        else:
            self.dump("final")

    def evaluate(self,
        eval_fpath: T.Optional[T.Union[Path, str]] = None,
        *,
        force: bool = False):

        eval_fpath = Path(eval_fpath)

        if eval_fpath.exists() and not force:
            logging.warning(f"Evaluation file (\"{eval_fpath}\") "\
                "exists already, skipping evaluation")
            return

        records = self.evaluator()

        records = {key: float(value) for key, value in records.items()}
        with open(eval_fpath, "w") as f:
            pyaml.dump(records, f, sort_keys=False)


    def _trainer_output(self, name: str = ""):
        return Path(self.trainer.out, name)

    def dump(self, suffix):
        self._check_attr("only_eval")
        if self.only_eval:
            return

        clf_file = self._trainer_output(f"clf_{suffix}.npz")
        logging.info(f"Storing classifier weights to {clf_file}")
        save_npz(clf_file, self.clf)

    def save_meta_info(self, meta_folder: str = "meta"):
        self._check_attr("config")

        folder = self._trainer_output(meta_folder)
        folder.mkdir(parents=True, exist_ok=True)

        with open(folder / "args.yml", "w") as f:
            pyaml.dump(self.config, f, sort_keys=True)




from cvfinetune.finetuner.mixins.dataset import _DatasetMixin
from cvfinetune.finetuner.mixins.classifier import _ClassifierMixin
from cvfinetune.finetuner.mixins.model import _ModelMixin
from cvfinetune.finetuner.mixins.optimizer import _OptimizerMixin
from cvfinetune.finetuner.mixins.iterator import _IteratorMixin
from cvfinetune.finetuner.mixins.trainer import _TrainerMixin


__all__ = [
	"_DatasetMixin",
	"_ClassifierMixin",
	"_ModelMixin",
	"_OptimizerMixin",
	"_IteratorMixin",
	"_TrainerMixin",
]



import gc
import logging
import threading


from chainer.training import extension
from chainer.training import trigger as trigger_module

class ManualGCCollect(extension.Extension):

	SLEEP = 3

	def __init__(self, trigger=(1, "iteration")):
		super().__init__()

		self._trigger = trigger_module.get_trigger(trigger)

		self.thread = threading.Thread(target=self.work)

		self.stop = threading.Event()
		self.trigger_gc = threading.Event()

	def work(self):
		logging.info("GC Thread working ...")

		while True:
			if self.stop.is_set():
				break

			if not self.trigger_gc.wait(self.SLEEP):
				continue

			self.trigger_gc.clear()
			gc.collect()

		logging.info("GC Thread finished")

	def initialize(self, trainer):
		logging.info("Starting GC Thread ...")
		self.thread.start()

	def __call__(self, trainer):
		if not self._trigger(trainer):
			return
		self.trigger_gc.set()


	def finalize(self):
		logging.info("Finilizing GC Thread")
		self.stop.set()
		self.thread.join()

	def on_error(self, trainer, exc, tb):
		logging.info("Error occured, stopping GC Thread")
		self.stop.set()
		self.thread.join()



import logging
import munch
import os
import re
import sys
import typing as T

from sacred import Experiment as BaseExperiment
from sacred.observers import MongoObserver
from sacred.utils import apply_backspaces_and_linefeeds

from pathlib import Path
from urllib.parse import quote_plus

def progress_bar_filter(text,
	escape=re.compile(r"\x1B\[([0-?]*[ -/]*[@-~])"),
	line_contents=re.compile(r".*(total|validation|this epoch|Estimated time|\d+ iter, \d+ epoch|\d+ \/ \d+ iteration).+\n"),
	tqdm_progress = re.compile(r"\n? *\d+\%\|.+\n?")):

	""" Filters out the progress bar of chainer """

	_text = apply_backspaces_and_linefeeds(text)

	_text = escape.sub("", _text)
	_text = line_contents.sub("", _text)
	_text = tqdm_progress.sub("", _text)
	_text = re.sub(r"\n *\n*", "\n", _text)

	return _text


class Experiment(BaseExperiment):

	ENV_KEYS: munch.Munch = munch.munchify(dict(
		USER_NAME="MONGODB_USER_NAME",
		PASSWORD="MONGODB_PASSWORD",
		DB_NAME="MONGODB_DB_NAME",

		HOST="MONGODB_HOST",
		PORT="MONGODB_PORT",
	))

	def __init__(self, *args,
		config: dict = {},
		host: T.Optional[str] = None,
		port: T.Optional[int] = None,
		no_observe: bool = False,
		output_filter: T.Callable = progress_bar_filter,
		**kwargs):

		if kwargs.get("base_dir") is None:
			base_dir = Path(sys.argv[0]).resolve().parent
			logging.info(f"Base experiment directory: {base_dir}")
			kwargs["base_dir"] = str(base_dir)

		super(Experiment, self).__init__(*args, **kwargs)

		if no_observe:
			return

		self.logger = logging.getLogger()
		self.captured_out_filter = output_filter

		creds = Experiment.get_creds()
		_mongo_observer = MongoObserver.create(
			url=Experiment.auth_url(creds, host=host, port=port),
			db_name=creds["db_name"],
		)

		self.observers.append(_mongo_observer)

		self.add_config(**config)

	def __call__(self, *args, **kwargs):
		return self._create_run()(*args, **kwargs)

	@classmethod
	def get_creds(cls):
		return dict(
			user=cls._get_env_key(cls.ENV_KEYS.USER_NAME),
			password=cls._get_env_key(cls.ENV_KEYS.PASSWORD),
			db_name=cls._get_env_key(cls.ENV_KEYS.DB_NAME),
		)

	@classmethod
	def auth_url(cls, creds, host="localhost", port=27017):
		host = host or cls.get_host()
		port = port or cls.get_port()
		logging.info(f"MongoDB host: {host}:{port}")

		url = "mongodb://{user}:{password}@{host}:{port}/{db_name}?authSource=admin".format(
			host=host, port=port, **creds)
		return url

	@classmethod
	def get_host(cls):
		return cls._get_env_key(cls.ENV_KEYS.HOST, default="localhost")

	@classmethod
	def get_port(cls):
		return cls._get_env_key(cls.ENV_KEYS.PORT, default=27017)

	@classmethod
	def _get_env_key(cls, key, default=None):
		return quote_plus(str(os.environ.get(key, default)))



__all__ = [
	"Experiment",
	"progress_bar_filter"
]



from chainer import functions as F


class SmoothedCrossEntropy:

	def __init__(self, n_classes: int, eps: float = 0.1):
		super(SmoothedCrossEntropy, self).__init__()
		self.n_classes = n_classes
		self.eps = eps

	def __call__(self, pred, gt, **kwargs):
		loss = F.softmax_cross_entropy(pred, gt, **kwargs)

		# -sum[ log( P(k) ) * U ]
		reg_loss = F.mean(F.sum(F.log_softmax(pred) / self.n_classes, axis=1))

		return (1-self.eps) * loss - self.eps * reg_loss


def smoothed_cross_entropy(N: int, eps: float = 0.1):
	return SmoothedCrossEntropy(n_classes=N, eps=eps)



import abc
import logging

from chainer import functions as F
from cvdatasets.utils import pretty_print_dict
from cvfinetune.training.loss import smoothed_cross_entropy
from functools import partial

from cvfinetune.finetuner.mixins.base import BaseMixin

class _ClassifierCreator:

    def __init__(self, cls, **kwargs):
        super().__init__()
        self.cls = cls
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        self.kwargs = dict(self.kwargs, **kwargs)
        return self.cls(*args, **self.kwargs)

class _ClassifierMixin(BaseMixin):
    """
        This mixin implements the wrapping of the backbone model around
        a classifier instance.
    """

    def __init__(self, *args,
                 classifier_cls,
                 classifier_kwargs: dict = {},
                 l1_loss: bool = False,
                 label_smoothing: float = 0.0,
                 **kwargs):

        super().__init__(*args, **kwargs)
        self._clf_creator = _ClassifierCreator(classifier_cls, **classifier_kwargs)

        self._l1_loss = l1_loss
        self._label_smoothing = label_smoothing


    def init_classifier(self, **kwargs):
        self._check_attr("model")
        self._check_attr("n_classes")

        self.clf = self._clf_creator(self.model,
                                     loss_func=self.loss_func,
                                     **kwargs)

        kwargs = self._clf_creator.kwargs
        logging.info(
            f"Wrapped the model around {type(self.clf).__name__}"
            f" with kwargs: {pretty_print_dict(kwargs)}"
        )

    @property
    def loss_func(self):
        if self._l1_loss:
            return F.hinge

        if self._label_smoothing > 0:
            assert self._label_smoothing < 1, "Label smoothing factor must be less than 1!"

            return smoothed_cross_entropy(self.n_classes, eps=self._label_smoothing)

        return F.softmax_cross_entropy



import abc
import numpy as np

from cvdatasets.dataset import AnnotationsReadMixin
from cvdatasets.dataset import IteratorMixin
from cvdatasets.dataset import TransformMixin
from cvdatasets.dataset import UniformPartMixin

class BaseDataset(TransformMixin, UniformPartMixin, AnnotationsReadMixin):
	"""Commonly used dataset constellation"""

	def __init__(self, *args, prepare, center_crop_on_val: bool = True, **kwargs):
		super().__init__(*args, **kwargs)
		raise NotImplementedError("YOU SHOULD NOT USE ME!")
		self.prepare = prepare

	def augment(self, im):
		if isinstance(im, list):
			im = np.array(im)

		if np.logical_and(0 <= im, im <= 1).all():
			im = im * 2 -1

		return im

	def transform(self, im_obj):
		im, parts, lab = im_obj.as_tuple()
		return self.prepare(im), lab + self.label_shift



import chainer
import logging
import numpy as np
import typing as T

from cvfinetune.finetuner import mixins

class DefaultFinetuner(
	mixins._ModelMixin,
	mixins._OptimizerMixin,
	mixins._ClassifierMixin,
	mixins._DatasetMixin,
	mixins._IteratorMixin,
	mixins._TrainerMixin):
	""" The default Finetuner gathers together the creations of all needed
	components and call them in the correct order

	"""

	def __init__(self, *args, config: dict = {}, gpu = [-1],
		         seed: T.Optional[int] = None, **kwargs):

		super().__init__(*args, **kwargs)
		self.init_random_state(seed=seed)
		self.init_experiment(config=config)

		self.gpu_config(gpu)
		self.read_annotations()

		self.init_model()
		self.init_datasets()
		self.init_iterators()

		self.init_classifier()
		self.load_weights()

		self.init_optimizer()
		self.init_updater()
		self.init_evaluator()


	def _check_attr(self, attr_name, msg=None):
		msg = msg or f"<{type(self).__name__}> {attr_name} attribute was not initialized!"
		assert hasattr(self, attr_name), msg

	def init_random_state(self, seed: T.Optional[int] = None):
		logging.info(f"Using seed: {seed}")
		self.rnd = np.random.RandomState(seed)

	def init_device(self):
		self.device = chainer.get_device(self.device_id)
		self.device.use()
		return self.device

	def gpu_config(self, devices):
		if -1 in devices:
			self.device_id = -1
		else:
			self.device_id = devices[0]

		device = self.init_device()
		logging.info(f"Using device {device}")
		return device




import wandb

from cvfinetune.training.extensions.reporters.base import BaseReport

class WandbReport(BaseReport):

	def reporter_enabled(self) -> bool:
		return True

	def log(self, key: str, value: float, step: int = 0) -> None:
		wandb.log({key: value}, step=step, commit=False)

	def __call__(self, trainer):
		# self.log will be called with commit=False,
		# so we need to call it once with commit=True
		super().__call__(trainer)
		wandb.log({}, commit=True)



import abc

from chainer import reporter
from chainer.training import trigger as trigger_module
from chainer.training.extension import Extension

class BaseReport(abc.ABC, Extension):
	def __init__(self, *, keys=None, trigger=(1, "epoch")):
		super().__init__()
		self._keys = keys
		self._trigger = trigger_module.get_trigger(trigger)

		self._init_summary()

	def __call__(self, trainer):
		if not self.reporter_enabled():
			return

		obs = trainer.observation
		keys = self._keys

		if keys is None:
			self._summary.add(obs)
		else:
			self._summary.add({k: obs[k] for k in keys if k in obs})

		if not self._trigger(trainer):
			return

		stats = self._summary.compute_mean()

		step = None
		if self._trigger.unit == "epoch":
			step = trainer.updater.epoch

		elif self._trigger.unit == "iteration":
			step = trainer.updater.iteration

		for name in stats:
			self.log(name, float(stats[name]), step=step)

		self._init_summary()

	def _init_summary(self):
		self._summary = reporter.DictSummary()

	@abc.abstractmethod
	def reporter_enabled(self) -> bool:
		pass

	@abc.abstractmethod
	def log(self, key: str, value: float, step: int = 0) -> None:
		pass





import logging
import warnings

from cvfinetune import utils
from cvfinetune.utils import mpi
from cvfinetune.finetuner.base import DefaultFinetuner
from cvfinetune.finetuner.mpi import MPIFinetuner

from cvdatasets.utils import pretty_print_dict

class FinetunerFactory:

	@classmethod
	def new(cls, *args, **kwargs):
		raise NotImplementedError("Use simple instance creation instead of {cls.__name__}.new()!")

	def __init__(self, *,
		         default=DefaultFinetuner,
		         mpi_tuner=MPIFinetuner,
		         **kwargs):
		super().__init__()

		if "mpi" in kwargs:
			kwargs.pop("mpi")
			warnings.warn("\"mpi\" is no longer supported. MPI checks are performed automatically.", category=DeprecationWarning)

		self.kwargs = kwargs
		self.tuner_cls = default

		if mpi.enabled() and mpi.chainermn_available():
			comm = mpi.new_comm("pure_nccl")
			msg1 = "MPI enabled. Creating NCCL communicator!"
			msg2 = f"Rank: {comm.rank}, IntraRank: {comm.intra_rank}, InterRank: {comm.inter_rank}"
			utils.log_messages(msg1, msg2)

			self["comm"] = comm
			self.tuner_cls = mpi_tuner

		logging.info(f"Using {self.tuner_cls.__name__} with arguments: {pretty_print_dict(self.kwargs)}")

	def __call__(self, opts, **kwargs):
		opt_kwargs = self.tuner_cls.extract_kwargs(opts)
		_kwargs = dict(self.kwargs, **kwargs, **opt_kwargs)
		return self.tuner_cls(config=opts.__dict__, **_kwargs)

	def get(self, key, default=None):
		return self.kwargs.get(key, default)

	def __getitem__(self, key):
		return self.kwargs[key]

	def __setitem__(self, key, value):
		self.kwargs[key] = value



import abc
import chainer
import logging

from chainer.optimizer_hooks import Lasso
from chainer.optimizer_hooks import WeightDecay
from chainer_addons.training import optimizer as new_optimizer
from chainer_addons.training.optimizer_hooks import SelectiveWeightDecay
from cvdatasets.utils import pretty_print_dict

from cvfinetune.finetuner.mixins.base import BaseMixin

def check_param_for_decay(param):
    return param.name != "alpha"

def enable_only_head(chain: chainer.Chain):
    if hasattr(chain, "enable_only_head") and callable(chain.enable_only_head):
        chain.enable_only_head()

    else:
        chain.disable_update()
        chain.fc.enable_update()

class _OptimizerCreator:

    def __init__(self, opt, **kwargs):
        super().__init__()

        self.opt = opt
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        if self.opt is None:
            return None

        kwargs = dict(self.kwargs, **kwargs)
        return new_optimizer(self.opt, *args, **kwargs)

class _OptimizerMixin(BaseMixin):

    def __init__(self, *args,
                 optimizer: str,
                 learning_rate: float = 1e-3,
                 weight_decay: float = 5e-4,
                 eps: float = 1e-2,
                 only_head: bool = False,
                 **kwargs):

        super().__init__(*args, **kwargs)

        optimizer_kwargs = dict(decay=0, gradient_clipping=False)

        if optimizer in ["rmsprop", "adam"]:
            optimizer_kwargs["eps"] = eps

        self._opt_creator = _OptimizerCreator(optimizer, **optimizer_kwargs)
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self._only_head = only_head


    def init_optimizer(self):
        """Creates an optimizer for the classifier """

        self._check_attr("clf")
        self._check_attr("_pooling")
        self._check_attr("_l1_loss")

        self.opt = self._opt_creator(self.clf, self.learning_rate)

        if self.opt is None:
            logging.warning("========= No optimizer was initialized! =========")
            return

        kwargs = self._opt_creator.kwargs
        logging.info(
            f"Initialized {type(self.opt).__name__} optimizer"
            f" with initial LR {self.learning_rate} and kwargs: {pretty_print_dict(kwargs)}"
        )

        self.init_regularizer()

        if self._only_head:
            logging.warning("========= Fine-tuning only classifier layer! =========")
            enable_only_head(self.clf)

    def init_regularizer(self, **kwargs):

        if self.weight_decay <= 0:
            return

        if self._l1_loss:
            cls = Lasso

        elif self._pooling == "alpha":
            cls = SelectiveWeightDecay
            kwargs["selection"] = check_param_for_decay

        else:
            cls = WeightDecay

        logging.info(f"Adding {cls.__name__} ({self.weight_decay:e})")
        self.opt.add_hook(cls(self.weight_decay, **kwargs))



import mpi4py.MPI as MPI

try:
	import chainermn
except Exception as e: #pragma: no cover
	_CHAINERMN_AVAILABLE = False #pragma: no cover
else:
	_CHAINERMN_AVAILABLE = True



def chainermn_available(strict: bool = True) -> bool:
	if strict:
		assert _CHAINERMN_AVAILABLE, "Distributed training is not possible!"

	return _CHAINERMN_AVAILABLE

def enabled() -> bool:
	return MPI.COMM_WORLD.Get_size() > 1

def new_comm(comm_type: str = "pure_nccl"):
	return chainermn.create_communicator("pure_nccl")



from cvfinetune.finetuner.base import DefaultFinetuner
from cvfinetune.finetuner.factory import FinetunerFactory
from cvfinetune.finetuner.mpi import MPIFinetuner

__all__ = [
	"FinetunerFactory",
	"DefaultFinetuner",
	"MPIFinetuner",
]



import logging
import numpy as np
import os
import typing as T
import warnings
import yaml

from cvargparse import BaseParser
from cvdatasets.utils import read_info_file
from functools import wraps

from pathlib import Path


WARNING = """Could not find default info file \"{}\". """ + \
"""Some arguments (dataset, parts etc.) are not restraint to certain choices! """ + \
"""You can set <DATA> environment variable to change the default info file location."""

DEFAULT_INFO_FILE = os.environ.get("DATA")
def get_info_file():

	if DEFAULT_INFO_FILE is not None and os.path.isfile(DEFAULT_INFO_FILE):
		return read_info_file(DEFAULT_INFO_FILE)
	else:
		warnings.warn(WARNING.format(DEFAULT_INFO_FILE))
		return None


def parser_extender(extender):

	@wraps(extender)
	def inner(parser, *args, **kwargs):
		assert isinstance(parser, BaseParser), \
			"Parser should be an BaseParser instance!"

		extender(parser, *args, **kwargs)

		return parser

	return inner


def populate_args(args, ignore = None, replace = {}, fc_params = []):

	args.debug = False

	assert args.load is not None, "--load argument missing!"

	model_path = Path(args.load)

	args_path = model_path.parent / "meta" / "args.yml"

	assert args_path.exists(), f"Couldn't find args file \"{args_path}\""

	logging.info(f"Reading arguments from \"{args_path}\"")

	with open(args_path) as f:
		dumped_args = yaml.safe_load(f)

	for key, value in dumped_args.items():
		if (ignore is not None and key in ignore) or getattr(args, key, None) == value:
			continue

		if key in ["logfile", "loglevel"]:
			continue

		old_value = getattr(args, key, None)
		if key in replace:
			value = replace[key].get(value, value)

		logging.debug(f"Setting \"{key}\" to {value} (originally was {'missing' if old_value is None else old_value})")

		setattr(args, key, value)

	# get the correct number of classes
	args.n_classes = get_n_classes(args.load, args.load_path,
		fc_params=fc_params)

	if args.n_classes <= 0:
		raise KeyError("Could not find number of classes!")

def get_n_classes(weights_file: str, load_path: str = "",
	*, fc_params: T.Tuple[str] = ()) -> int:
	"""
		Searches for the classification layer identified by
		names in fc_params.
	"""

	weights = np.load(weights_file)
	prefixes = [""]

	if load_path != "":
		prefixes.append(load_path)

	for prefix in prefixes:
		for name in fc_params:
			key = f"{prefix}{name}"
			if key not in weights:
				continue
			return weights[key].shape[0]

	return -1






import chainermn
import logging
import warnings

from chainermn import scatter_dataset as scatter

from cvfinetune.finetuner.base import DefaultFinetuner

class MPIFinetuner(DefaultFinetuner):

	def __init__(self, *args, comm, **kwargs):
		self.comm = comm
		super(MPIFinetuner, self).__init__(*args, **kwargs)

	@property
	def mpi_main_process(self):
		return not (self.comm is not None and self.comm.rank != 0)

	@property
	def no_observe(self):
		return self.no_sacred or not self.mpi_main_process

	def check_mpi(self):
		if self.comm is None:
			msg = "Using MPIFinetuner without setting a communicator!"
			warnings.warn(msg)
			logging.warn(msg)
			return False

		return True

	def gpu_config(self, devices):

		if not self.check_mpi():
			return super(MPIFinetuner, self).gpu_config(devices)

		if len(devices) == 1:
			self.device_id = devices[0] + self.comm.intra_rank
		else:
			self.device_id = devices[self.comm.rank]

		device = self.init_device()
		ranks = f"{self.comm.rank} | {self.comm.intra_rank} | {self.comm.inter_rank}"
		logging.info(f"Node with ranks {ranks} assigned to {device}")
		return device


	def scatter_datasets(self):
		if self.check_mpi():
			self.train_data = scatter(self.train_data, self.comm)
			self.val_data = scatter(self.val_data, self.comm)
		else:
			logging.warn("Data scattering was not Possible!")


	def init_datasets(self, *args, **kwargs):

		if self.mpi_main_process:
			super(MPIFinetuner, self).init_datasets(*args, **kwargs)
		else:
			self.train_data, self.val_data = None, None

		self.scatter_datasets()

	def init_optimizer(self):
		super(MPIFinetuner, self).init_optimizer()

		if not self.check_mpi():
			return

		self.opt = chainermn.create_multi_node_optimizer(self.opt, self.comm)

	def init_evaluator(self):
		super(MPIFinetuner, self).init_evaluator()

		if not self.check_mpi():
			return

		self.evaluator = chainermn.create_multi_node_evaluator(
			self.evaluator, self.comm)

	def run(self, trainer_cls, opts, *args, **kwargs):
		if not self.mpi_main_process:
			kwargs["no_observe"] = True
			opts.no_snapshot = True
			opts.no_progress = True
			self.evaluator._progress_bar = False

		super(MPIFinetuner, self).run(trainer_cls, opts, *args, **kwargs)



import abc
import inspect


class BaseMixin(abc.ABC):

	def _after_init_check(self):
		pass

	@classmethod
	def extract_kwargs(cls, opts) -> dict:

		kwargs = {}

		for klass in cls.mro():
			sig = inspect.signature(klass.__init__)
			for attr, param in sig.parameters.items():
				if param.kind is not inspect.Parameter.KEYWORD_ONLY:
					continue

				if param.name in kwargs:
					continue

				if hasattr(opts, param.name):
					value = getattr(opts, param.name)
					kwargs[param.name] = value
		return kwargs


if __name__ == '__main__':

	from collections import namedtuple
	class Foo(BaseMixin):

		@classmethod
		def extract_kwargs(cls, opts) -> dict:
			return super().extract_kwargs(opts)

		def __init__(self, *args, foo, bar=0, **kwargs):
			super().__init__(*args, **kwargs)
			self.foo = foo
			self.bar = bar


	class Bar(BaseMixin):
		@classmethod
		def extract_kwargs(cls, opts) -> dict:
			return super().extract_kwargs(opts)

		def __init__(self, *args, bar2=-1, **kwargs):
			super().__init__(*args, **kwargs)
			self.bar2 = bar2


	class Final(Bar, Foo):


		def __init__(self, *args, beef=-1, **kwargs):
			super().__init__(*args, **kwargs)
			self.beef = beef

		def __repr__(self):
			return str(self.__dict__)


	_Opts = namedtuple("_Opts", "foo foo2 bar bar2 beef1")

	opts = _Opts(1,2,3, -4, "hat")
	kwargs = Final.extract_kwargs(opts)

	print(opts, Final(**kwargs))



import abc
import chainer
import logging
import warnings

from chainer import functions as F
from chainer_addons.models import PrepareType
from chainercv2.models import model_store
from cvdatasets.dataset.image import Size
from cvmodelz.models import ModelFactory
from functools import partial
from pathlib import Path
from typing import Tuple

from cvfinetune.finetuner.mixins.base import BaseMixin

class _ModelMixin(BaseMixin):
    """
        This mixin is responsible for model selection, model and optimizer creation,
        and model weights loading.
    """

    def __init__(self, *args,
                 model_type: str,
                 model_kwargs: dict = {},
                 pooling: str = "g_avg",

                 prepare_type: str = "model",
                 center_crop_on_val: bool = True,
                 swap_channels: bool = False,

                 load: str = None,
                 weights: str = None,
                 load_path: str = "",
                 load_strict: bool = False,
                 load_headless: bool = False,
                 pretrained_on: str = "imagenet",

                 from_scratch: bool = False,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.model_type = model_type
        self.model_kwargs = model_kwargs

        self._center_crop_on_val = center_crop_on_val
        self._swap_channels = swap_channels

        if model_type.startswith("chainercv2"):
            if prepare_type != "chainercv2":
                msg = f"Using chainercv2 model, but prepare_type was set to \"{prepare_type}\". "
                "Setting it to \"chainercv2\"!"
                warnings.warn(msg)
            prepare_type = "chainercv2"

        self._prepare_type = prepare_type
        self._pooling = pooling

        self._load = load
        self._weights = weights
        self._from_scratch = from_scratch
        self._load_path = load_path
        self._load_strict = load_strict
        self._load_headless = load_headless
        self._pretrained_on = pretrained_on


    def init_model(self):
        """creates backbone CNN model. This model is wrapped around the classifier later"""

        self._check_attr("input_size")

        self.model = self.new_model()

        logging.info(
            f"Created {type(self.model).__name__} model "
            f" with \"{self._prepare_type}\" prepare function."
        )


    def load_weights(self) -> None:

        self._check_attr("clf")
        self._check_attr("n_classes")

        finetune, weights = self._get_loader()

        self.clf.load(weights,
            n_classes=self.n_classes,
            finetune=finetune,

            path=self._load_path,
            strict=self._load_strict,
            headless=self._load_headless
        )

        self.clf.cleargrads()

        feat_size = self.model.meta.feature_size

        if hasattr(self.clf, "output_size"):
            feat_size = self.clf.output_size

        ### TODO: handle feature size!

        logging.info(f"Part features size after encoding: {feat_size}")



    @property
    def prepare_type(self):
        return PrepareType[self._prepare_type]

    @property
    def prepare(self):
        return partial(self.prepare_type(self.model),
            swap_channels=self._swap_channels,
            keep_ratio=self._center_crop_on_val)

    def new_model(self, **kwargs):
        return ModelFactory.new(self.model_type,
            input_size=self.input_size,
            **self.model_kwargs, **kwargs)

    @property
    def model_info(self):
        return self.data_info.MODELS[self.model_type]



    def _get_loader(self) -> Tuple[bool, str]:

        if self._from_scratch:
            logging.info(f"Training a {type(self.model).__name__} model from scratch!")
            return None, None

        if self._load:
            weights = self._load
            logging.info(f"Loading already fine-tuned weights from \"{weights}\"")
            return False, weights

        elif self._weights:
            weights = self._weights
            logging.info(f"Loading custom fine-tuned weights from \"{weights}\"")
            return True, weights

        else:
            weights = self._default_weights
            logging.info(f"Loading default fine-tuned weights from \"{weights}\"")
            return True, weights

    @property
    def _default_weights(self):
        if self.model_type.startswith("chainercv2"):
            model_name = self.model_type.split(".")[-1]
            return model_store.get_model_file(
                model_name=model_name,
                local_model_store_dir_path=str(Path.home() / ".chainer" / "models"))

        else:
            ds_info = self.data_info
            model_info = self.model_info

            base_dir = Path(ds_info.BASE_DIR)
            weights_dir = base_dir / ds_info.MODEL_DIR / model_info.folder

            weights = model_info.weights
            assert self._pretrained_on in weights, \
                f"Weights for \"{self._pretrained_on}\" pre-training were not found!"

            return str(weights_dir / weights[self._pretrained_on])




from cvfinetune.training.extensions.reporters.base import BaseReport

class SacredReport(BaseReport):
	def __init__(self, *args, ex, **kwargs):
		self.ex = ex
		super().__init__(*args, **kwargs)

	def reporter_enabled(self) -> bool:
		return None not in [self.ex, self.ex.current_run]

	def log(self, key: str, value: float, step: int = 0) -> None:
		self.ex.log_scalar(key, value, step=step)



import abc
import logging
import typing as T

from cvdatasets.utils import new_iterator

from cvfinetune.finetuner.mixins.base import BaseMixin

class _IteratorMixin(BaseMixin):

    def __init__(self,
                 *args,
                 batch_size: int = 32,
                 n_jobs: int = 1,
                 use_threads: bool = False,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._batch_size = batch_size
        self._n_jobs = n_jobs
        self._use_threads = use_threads


    def new_iterator(self, ds, **kwargs):
        if hasattr(ds, "new_iterator"):
            return ds.new_iterator(**kwargs)
        else:
            return new_iterator(ds, **kwargs)

    def init_iterators(self):
        """Creates training and validation iterators from training and validation datasets"""

        self._check_attr("val_data")
        self._check_attr("train_data")

        kwargs = dict(
            n_jobs=self._n_jobs,
            batch_size=self._batch_size,
            use_threads=self._use_threads,
        )

        self.train_iter, _ = self.new_iterator(self.train_data,
                                               **kwargs)

        self.val_iter, _ = self.new_iterator(self.val_data,
                                             repeat=False, shuffle=False,
                                             **kwargs)



import abc
import logging
import typing as T
import chainer

from cvdatasets import AnnotationType
from cvdatasets import AnnotationArgs
from cvdatasets.dataset.image import Size

from cvfinetune.finetuner.mixins.base import BaseMixin

class _DatasetMixin(BaseMixin):
    """
        This mixin is responsible for annotation loading and for
        dataset and iterator creation.
    """

    def __init__(self,
                 *args,
                 data: str,
                 dataset: str,
                 dataset_cls: T.Type,
                 dataset_kwargs_factory: T.Optional[T.Callable] = None,

                 parts: str = "GLOBAL",
                 label_shift: int = 0,
                 input_size: int = 224,
                 feature_model: str = None,
                 part_input_size:  T.Optional[int] = None,
                 **kwargs):

        super().__init__(*args, **kwargs)
        self.annot = None
        self.info_file = data
        self.dataset_name = dataset
        self.part_type = parts
        self.feature_model = feature_model
        self.dataset_cls = dataset_cls
        self.dataset_kwargs_factory = dataset_kwargs_factory

        if part_input_size is None:
            part_input_size = input_size

        self.input_size = Size(input_size)
        self.part_input_size = Size(part_input_size)

        self._label_shift = label_shift


    def read_annotations(self):
        """
            Reads annotations and creates annotation instance,
            which holds important infos about the dataset
        """
        args = AnnotationArgs(
            self.info_file,
            self.dataset_name,
            self.part_type,
            self.feature_model
        )
        self.annot = AnnotationType.new_annotation(args, load_strict=False)
        self.dataset_cls.label_shift = self._label_shift

    def init_datasets(self):
        self._check_attr("prepare")
        self._check_attr("_center_crop_on_val")

        logging.info(" | ".join([
            f"Image input size: {self.input_size}",
            f"Parts input size: {self.part_input_size}",
        ]))

        self.train_data = self.new_dataset("train")
        self.val_data = self.new_dataset("test")


    @property
    def n_classes(self):
        return self.ds_info.n_classes + self._label_shift

    @property
    def data_info(self):
        assert self.annot is not None, "annot attribute was not set!"
        return self.annot.info

    @property
    def ds_info(self):
        return self.data_info.DATASETS[self.dataset_name]

    def new_dataset(self, subset: str):
        """Creates a dataset for a specific subset and certain options"""
        if self.dataset_kwargs_factory is not None and callable(self.dataset_kwargs_factory):
            kwargs = self.dataset_kwargs_factory(subset)
        else:
            kwargs = dict()

        kwargs = dict(kwargs,
            subset=subset,
            dataset_cls=self.dataset_cls,
            prepare=self.prepare,
            size=self.input_size,
            part_size=self.part_input_size,
            center_crop_on_val=self._center_crop_on_val,
        )


        ds = self.annot.new_dataset(**kwargs)
        logging.info(f"Loaded {len(ds)} images")
        return ds


    def profile_images(self, train: bool = True):
        ds = self.train_data if train else self.val_data
        subset_name = "training" if train else "validation"

        if isinstance(ds, chainer.datasets.SubDataset):
            ds = ds._dataset


        logging.info(f"Profiling {subset_name} image processing: ")
        with ds.enable_img_profiler():
            ds[self.rnd.randint(len(ds))]





from cvfinetune.training.extensions.gc_collect import ManualGCCollect
from cvfinetune.training.extensions.reporters.sacred import SacredReport
from cvfinetune.training.extensions.reporters.wandb import WandbReport
